#include <mono/jit/jit.h>
#include <mono/metadata/object.h>
#include <mono/metadata/environment.h>
#include <mono/metadata/assembly.h>
#include <mono/metadata/debug-helpers.h>
#include <mono/metadata/threads.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>


/*
 * Compile with: 
 * 	gcc -Wall -o test-invoke test-invoke.c `pkg-config --cflags --libs mono` -lm
 */



MonoMethod* GetMethod(MonoClass *klass, const char *name, int args)
{
    MonoMethod *m;
    gpointer   iter;

    // retrieve the method
    iter = NULL;
    while ((m = mono_class_get_methods(klass, &iter))) {
        if (strcmp (mono_method_get_name(m), name) == 0) {
            // Check the number of args:
            MonoMethodSignature *sig = mono_method_signature(m);
            if (mono_signature_get_param_count (sig) == args) {
                return m;
            }
        }
    }
    return NULL;
}



MonoClass* GetClass(MonoAssembly *assembly, const char* namespace, const char* className)
{
    MonoClass *klass;
    
        
    klass = mono_class_from_name(mono_assembly_get_image(assembly), namespace, className);
    if (!klass) {
        printf("Can't find %s.%s in assembly %s\n", namespace, className, mono_image_get_filename(mono_assembly_get_image(assembly)));
        exit(1);
    }
    
    return klass;
}



MonoAssembly* GetAssembly(const char* fileName)
{
    MonoAssembly  *assembly;
    MonoImageOpenStatus  status;
    
    
    assembly = mono_assembly_open(fileName, &status);
    if (!assembly) {
        printf("Can't load assembly %s.\n", fileName);
        exit(2);
    }
    
    return assembly;
}



int main()
{
    MonoDomain   *domain;
    MonoAssembly *assemblyBroadcast, *assemblyUtils; 
    MonoClass    *classBroadcast, *classUtils, *classThread;
    MonoObject   *objBroadcast;
    MonoMethod   *clientAttachBroadcast, *clientDetachBroadcast, *getLECServerList, *Sleep, *getBroadcastData;
    long          clientID;
	

    // creo il dominio
    domain = mono_jit_init("LECBroadcast.dll");
    
    // classe Thread per gestire la Sleep 
    // (in Mono embedding i segnali non vengono inoltrati al sistema quindi non posso usare la normale sleep Unix)
    classThread = mono_class_from_name(mono_get_corlib(), "System.Threading", "Thread");
    // puntatore alla funzione Sleep
    if ((Sleep = GetMethod(classThread, "Sleep", 1)) == NULL) {
        printf("Can't find Sleep in class Thread\n");
        exit(1);
    }
    
    // carico la classe Utils contenuta in LECUtils.dll
    assemblyUtils = GetAssembly("LECUtils.dll");
    classUtils = GetClass(assemblyUtils, "LanmarkControls.LECUtils", "Utils");
    
    // carico la classe Broadcast contenuta in LECBroadcast.dll
    assemblyBroadcast = GetAssembly("LECBroadcast.dll");
    classBroadcast = GetClass(assemblyBroadcast, "LanmarkControls.LECBroadcast", "Broadcast");
    
    // istanzio un oggetto broadcast
    printf("Creating BROADCAST object.\n");
    objBroadcast = mono_object_new(domain, classBroadcast);
    // eseguo il costruttore di default
    mono_runtime_object_init(objBroadcast);
    
    // puntatore alla funzione clientAttachBroadcast
    if ((clientAttachBroadcast = GetMethod(classBroadcast, "clientAttachBroadcast", 4)) == NULL) {
        printf("Can't find clientAttachBroadcast in class Broadcast\n");
        exit(1);
    }
    else {  // se la funzione esiste, la eseguo
        MonoObject *monoAttachResult;
        gpointer   args [4];
        long       port = 11000;
        long       resultAttach;
        
        // preparo i parametri da passare alla funzione
        args[0] = mono_string_new(domain, "224.168.100.2");
        args[1] = mono_string_new(domain, "0.0.0.0");
        args[2] = &port;
        args[3] = &clientID;
        // invoco la funzione
        monoAttachResult = mono_runtime_invoke(clientAttachBroadcast, objBroadcast, args, NULL);
        // converto il valore restituito
        resultAttach = *(long int*)mono_object_unbox(monoAttachResult);
        
        if (resultAttach != 0) {
            printf("   clientAttachBroadcast() FAILED. %ld\n", resultAttach);
            return 1;
        }
        
        printf("   clientAttachBroadcast() Suceeded. clientID=%ld\n", clientID);
    }
    
    // puntatore alla funzione getLECServerList
    if ((getLECServerList = GetMethod(classBroadcast, "getLECServerList", 3)) == NULL) {
        printf("Can't find getLECServerList in class Broadcast\n");
        return 1;
    }
    else {  // se la funzione esiste
        MonoObject *monoServerListResult;
        gpointer   args[3];
        gpointer   argsSleep[1];
        long       sleep = 1000;
        long       serverCount = 0;
        long       resultServerList;
        int        i;
        MonoString *serverXML;
        
        // parametri da passare alla funzione getLECServerList
        serverXML = mono_string_new(domain, " ");
        args[0] = &clientID;
        args[1] = &serverCount;
        args[2] = &serverXML;
        // parametro per la funzione Sleep
        argsSleep[0] = &sleep;
        
        // eseguo il polling per 10 secondi alla ricerca delle schede
        printf("Polling... ");
        for (i = 0; i < 10; i++) {
            // invoco la funzione getLECServerList
            monoServerListResult = mono_runtime_invoke(getLECServerList, objBroadcast, args, NULL);
            resultServerList = *(long*)mono_object_unbox(monoServerListResult);
            if (resultServerList != 0) {
                printf("   getLECServerList() FAILED. %ld\n", resultServerList);
                return 1;
            }

            if (Sleep != NULL) {
                gpointer   argsSleep[1];
                int        sleep = 1000;
                
                // preparo il parametro per la funzione
                argsSleep[0] = &sleep;
                // invoco la funzione Sleep
                mono_runtime_invoke(Sleep, NULL, argsSleep, NULL);
            }
        }
        printf("\nPoll found %ld target(s)\n", serverCount);
        
        
        if (serverCount <= 0)   // se non trovo nessuna scheda esco
            exit(0);
        
        
        char* addresses[serverCount];   // indirizzi IP schede
        int   ports[serverCount];       // porta schede
        
        // ricavo indirizzi e porte di ogni scheda rilevata
        for (i = 0; i < serverCount; i++) {
            MonoString  *name;
            MonoString  *data;
            MonoMethod  *ReadFromXML;
            gpointer     argsReadFromXML[5];
            
            
            // Ricavo il valore dell'attributo XML 'name' dallelemento LEC di ServerList
            
            // puntatore alla funzione ReadFromXML
            if ((ReadFromXML = GetMethod(classUtils, "ReadFromXML", 5)) == NULL) {
                printf("Can't find ReadFromXML in class Utils\n");
                exit(1);
            }
            
            // preparo i parametri da passare alla funzione
            name = mono_string_new(domain, " ");
            argsReadFromXML[0] = serverXML;
            argsReadFromXML[1] = &i;
            argsReadFromXML[2] = mono_string_new(domain, "LEC");
            argsReadFromXML[3] = mono_string_new(domain, "name");
            argsReadFromXML[4] = &name;
            // invoco la funzione ReadFromXML
            mono_runtime_invoke(ReadFromXML, NULL, argsReadFromXML, NULL);
            
            // recupero il puntatore alla funzione getBroadcastData
            // N.B.: la classe Broadcast contiene due funzioni con lo
            //       stesso numero di parametri; quella che serve a noi
            //       e' la seconda, quindi ho dovuto saltare la prima...
            MonoMethod *m;
            gpointer    iter;
            iter = NULL;
            getBroadcastData = NULL;
            int n = 0;
            while ((m = mono_class_get_methods(classBroadcast, &iter))) {
                if (strcmp (mono_method_get_name(m), "getBroadcastData") == 0) {
                    n++;
                    if (n == 2) {
                        getBroadcastData = m;
                        break;
                    }
                }
            }
            
            if (getBroadcastData == NULL) {
                printf("Can't find getBroadcastData in class Broadcast\n");
                exit(1);
            }
            
            MonoObject   *monoResultBroadcastData;
            gpointer      argsBroadcastData[4];
            long int      dataType;
            unsigned int  err;
            
            // preparo i parametri da passare alla funzione
            data = mono_string_new(domain, " ");
            dataType = 0x00000007;
            argsBroadcastData[0] = &clientID;
            argsBroadcastData[1] = name;
            argsBroadcastData[2] = &dataType;
            argsBroadcastData[3] = &data;
        
            err = 1;
            while (err != 0)
            {
                // invoco la funzione getBroadcastData
                monoResultBroadcastData = mono_runtime_invoke(getBroadcastData, objBroadcast, argsBroadcastData, NULL);
                // converto il valore restituito
                err = *(unsigned int*)mono_object_unbox(monoResultBroadcastData);
                if (err == 0) {
                    printf("\nSUCCESS getBroadcastData('%s', 0x00000007)\n", mono_string_to_utf8(name));
                }
                else {
                    printf(".");
                    
                    if (Sleep != NULL) {
                        gpointer   argsSleep[1];
                        int        sleep = 1000;
                        
                        // preparo i parametri per la funzione
                        argsSleep[0] = &sleep;
                        // invoco la funzione Sleep
                        mono_runtime_invoke(Sleep, NULL, argsSleep, NULL);
                    }
                }
            }
            
            printf("%s\n", mono_string_to_utf8(data));
            
            // preparo i parametri per la funzione
            dataType = 0x00000001;
            argsBroadcastData[0] = &clientID;
            argsBroadcastData[1] = name;
            argsBroadcastData[2] = &dataType;
            argsBroadcastData[3] = &data;

            err = 1;
            while (err != 0)
            {
                // invoco la funzione getBroadcastData
                monoResultBroadcastData = mono_runtime_invoke(getBroadcastData, objBroadcast, argsBroadcastData, NULL);
                // converto il valore restituito
                err = *(unsigned int*)mono_object_unbox(monoResultBroadcastData);
                if (err == 0) {
                    printf("\nSUCCESS getBroadcastData('%s', 0x00000001)\n", mono_string_to_utf8(name));
                }
                else {
                    printf(".");
                    
                    if (Sleep != NULL) {
                        gpointer   argsSleep[1];
                        int        sleep = 1000;
                        
                        // preparo i parametri per la funzione
                        argsSleep[0] = &sleep;
                        // invoco la funzione Sleep
                        mono_runtime_invoke(Sleep, NULL, argsSleep, NULL);
                    }
                }
            }
            
            printf("%s\n\n", mono_string_to_utf8(data));
            
            // estraggo dall'XML indirizzo IP e porta
            
            // preparo i parametri per la funzione
            i = 0;
            MonoString *addressIP;
            argsReadFromXML[0] = data;
            argsReadFromXML[1] = &i;
            argsReadFromXML[2] = mono_string_new(domain, "IP");
            argsReadFromXML[3] = &i;
            argsReadFromXML[4] = &addressIP;
            // invoco la funzione readFromXML
            mono_runtime_invoke(ReadFromXML, NULL, argsReadFromXML, NULL);
            
            strcpy(addresses[i], mono_string_to_utf8(addressIP));
            
            // preparo i parametri per la funzione
            i = 0;
            MonoString *clientPort;
            argsReadFromXML[0] = data;
            argsReadFromXML[1] = &i;
            argsReadFromXML[2] = mono_string_new(domain, "Port");
            argsReadFromXML[3] = &i;
            argsReadFromXML[4] = &clientPort;
            // invoco la funzione readFromXML
            mono_runtime_invoke(ReadFromXML, NULL, argsReadFromXML, NULL);
            
            ports[i] = atoi(mono_string_to_utf8(clientPort));
        }
        
        // invio ad ogni scheda un semplice Job che marca un quadrato
        
        for (i = 0; i < serverCount; i++)  {
            MonoAssembly *assemblySession; 
            MonoClass    *classSession;
            MonoObject   *objSession;
            MonoMethod   *loginSession, *sendStreamData, *logoutSession;
            
            // carico la classe Session contenuta in LECSession.dll
            assemblySession = GetAssembly("LECSession.dll");
            classSession = GetClass(assemblySession, "LanmarkControls.LECSession", "Session");
    
            // istanzio l'oggetto Session
            printf("Creating SESSION object.\n");
            objSession = mono_object_new(domain, classSession);
            // chiamo il costruttore di default
            mono_runtime_object_init(objSession);
            
            // puntatore alla funzione loginSession
            if ((loginSession = GetMethod(classSession, "loginSession", 6)) == NULL) {
                printf("Can't find loginSession in class Session\n");
                exit(1);
            }
            
            // puntatore alla funzione logoutSession
            if ((logoutSession = GetMethod(classSession, "logoutSession", 1)) == NULL) {
                printf("Can't find logoutSession in class Session\n");
                exit(1);
            }
            
            // puntatore alla funzione sendStreamData
            if ((sendStreamData = GetMethod(classSession, "sendStreamData", 2)) == NULL) {
                printf("Can't find sendStreamData in class Session\n");
                exit(1);
            }
            
            MonoObject *monoLoginResult;
            gpointer   argsLogin[6];
            unsigned long timeout = 10000;
            unsigned long resultLogin;
        
            // preparo i parametri per la funzione
            argsLogin[0] = mono_string_new(domain, "0.0.0.0");
            argsLogin[1] = mono_string_new(domain, addresses[i]);
            argsLogin[2] = &(ports[i]);
            argsLogin[3] = mono_string_new(domain, "");
            argsLogin[4] = mono_string_new(domain, "");
            argsLogin[5] = &timeout;
            // invoco la funzione loginSession
            monoLoginResult = mono_runtime_invoke(loginSession, objSession, argsLogin, NULL);
            // converto il risultato
            resultLogin = *(unsigned long*)mono_object_unbox(monoLoginResult);
        
            if (resultLogin == 0) { // se ho eseguito il login                
                MonoObject    *monoSendResult;
                gpointer       argsSend[2];
                unsigned long  timeoutSend = 10000;
                unsigned long  resultSend;
                
                
                printf("\nLogged in to %s\n", addresses[i]);
                printf("\nSending sample job...\n");
                
                // preparo i parametri per la funzione
                argsSend[0] = mono_string_new(domain, "<Data type='JobData' rev='1.0'><Set id='JumpSpeed'>15,30</Set><Set id='MarkSpeed'>15,10</Set><JumpAbs>-5000,-5000</JumpAbs><MarkAbs>-5000,5000</MarkAbs><MarkAbs>5000,5000</MarkAbs><MarkAbs>5000,-5000</MarkAbs><MarkAbs>-5000,-5000</MarkAbs></Data>");
                argsSend[1] = &timeoutSend;
                // invoco la funzione sendStreamData
                monoSendResult = mono_runtime_invoke(sendStreamData, objSession, argsSend, NULL);
                // converto il risultato
                resultSend = *(unsigned long*)mono_object_unbox(monoSendResult);
                if (resultSend == 0) {
                    printf("Job sent!\n");
                }
                else {
                    printf("ERROR %ld sending job!\n", resultSend);
                }
                
                gpointer argsLogout[1];
                
                // preparo i parametri per la funzione
                argsLogout[0] = &timeoutSend;
                // invoco la funzione logoutSession
                mono_runtime_invoke(logoutSession, objSession, argsLogout, NULL);
                printf("\nLogged out to %s\n", addresses[i]);
            }
            else {
                printf("\nERROR: failed to login to target: %s LecSession client\n", addresses[i]);
            }
        }
    }
    
    // puntatore alla funzione clientDetachBroadcast
    if ((clientDetachBroadcast = GetMethod(classBroadcast, "clientDetachBroadcast", 1)) == NULL) {
        printf("Can't find clientDetachBroadcast in class Broadcast\n");
        exit(1);
    }
    else {
        gpointer args[1];
        
        // preparo i parametri per la funzione
        args[0] = &clientID;
        // invoco la funzione clientDetachBroadcast
        mono_runtime_invoke(clientDetachBroadcast, objBroadcast, args, NULL);
    }
	
	mono_jit_cleanup(domain);
    
    return 0;
}

